@@ -1,15 +1,18 @@
+from __future__ import annotations
 # built-in
 from collections import defaultdict
 from itertools import zip_longest
+from typing import Any, Sequence, TypeVar
 
 # app
 from .base import Base as _Base, BaseSimilarity as _BaseSimilarity
+from .types import TestFunc, SimFunc
 
 
 try:
     import numpy
 except ImportError:
-    numpy = None
+    numpy = None  # type: ignore[assignment]
 
 
 __all__ = [
@@ -23,6 +26,7 @@ __all__ = [
     'jaro', 'jaro_winkler', 'strcmp95',
     'needleman_wunsch', 'gotoh', 'smith_waterman',
 ]
+T = TypeVar('T')
 
 
 class Hamming(_Base):
@@ -32,21 +36,29 @@ class Hamming(_Base):
 
     https://en.wikipedia.org/wiki/Hamming_distance
     """
-    def __init__(self, qval=1, test_func=None, truncate=False, external=True):
+
+    def __init__(
+        self,
+        qval: int = 1,
+        test_func: TestFunc | None = None,
+        truncate: bool = False,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.test_func = test_func or self._ident
         self.truncate = truncate
         self.external = external
 
-    def __call__(self, *sequences):
+    def __call__(self, *sequences: Sequence[object]) -> int:
         sequences = self._get_sequences(*sequences)
 
         result = self.quick_answer(*sequences)
         if result is not None:
+            assert isinstance(result, int)
             return result
 
         _zip = zip if self.truncate else zip_longest
-        return sum([not self.test_func(*es) for es in _zip(*sequences)])
+        return sum(not self.test_func(*es) for es in _zip(*sequences))
 
 
 class Levenshtein(_Base):
@@ -62,12 +74,18 @@ class Levenshtein(_Base):
     https://en.wikipedia.org/wiki/Levenshtein_distance
     TODO: https://gist.github.com/kylebgorman/1081951/9b38b7743a3cb5167ab2c6608ac8eea7fc629dca
     """
-    def __init__(self, qval=1, test_func=None, external=True):
+
+    def __init__(
+        self,
+        qval: int = 1,
+        test_func: TestFunc | None = None,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.test_func = test_func or self._ident
         self.external = external
 
-    def _recursive(self, s1, s2):
+    def _recursive(self, s1: Sequence[T], s2: Sequence[T]) -> int:
         # TODO: more than 2 sequences support
         if not s1 or not s2:
             return len(s1) + len(s2)
@@ -84,7 +102,7 @@ class Levenshtein(_Base):
         s = self(s1[:-1], s2[:-1])
         return min(d, s) + 1
 
-    def _cicled(self, s1, s2):
+    def _cycled(self, s1: Sequence[T], s2: Sequence[T]) -> int:
         """
         source:
         https://github.com/jamesturk/jellyfish/blob/master/jellyfish/_jellyfish.py#L18
@@ -92,6 +110,7 @@ class Levenshtein(_Base):
         rows = len(s1) + 1
         cols = len(s2) + 1
         prev = None
+        cur: Any
         if numpy:
             cur = numpy.arange(cols)
         else:
@@ -107,14 +126,15 @@ class Levenshtein(_Base):
                 cur[c] = min(edit, deletion, insertion)
         return cur[-1]
 
-    def __call__(self, s1, s2):
+    def __call__(self, s1: Sequence[T], s2: Sequence[T]) -> int:
         s1, s2 = self._get_sequences(s1, s2)
 
         result = self.quick_answer(s1, s2)
         if result is not None:
+            assert isinstance(result, int)
             return result
 
-        return self._cicled(s1, s2)
+        return self._cycled(s1, s2)
 
 
 class DamerauLevenshtein(_Base):
@@ -128,14 +148,26 @@ class DamerauLevenshtein(_Base):
         * substitution:  ABC -> ABE, ADC, FBC..
         * transposition: ABC -> ACB, BAC
 
+    If `restricted=False`, it will calculate unrestricted distance,
+    where the same character can be touched more than once.
+    So the distance between BA and ACB is 2: BA -> AB -> ACB.
+
     https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance
     """
-    def __init__(self, qval=1, test_func=None, external=True):
+
+    def __init__(
+        self,
+        qval: int = 1,
+        test_func: TestFunc | None = None,
+        external: bool = True,
+        restricted: bool = True,
+    ) -> None:
         self.qval = qval
         self.test_func = test_func or self._ident
         self.external = external
+        self.restricted = restricted
 
-    def _numpy(self, s1, s2):
+    def _numpy(self, s1: Sequence[T], s2: Sequence[T]) -> int:
         # TODO: doesn't pass tests, need improve
         d = numpy.zeros([len(s1) + 1, len(s2) + 1], dtype=int)
 
@@ -168,11 +200,52 @@ class DamerauLevenshtein(_Base):
 
         return d[len(s1) - 1][len(s2) - 1]
 
-    def _pure_python(self, s1, s2):
+    def _pure_python_unrestricted(self, s1: Sequence[T], s2: Sequence[T]) -> int:
+        """https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance
+        """
+        d: dict[tuple[int, int], int] = {}
+        da: dict[T, int] = {}
+
+        len1 = len(s1)
+        len2 = len(s2)
+
+        maxdist = len1 + len2
+        d[-1, -1] = maxdist
+
+        # matrix
+        for i in range(len(s1) + 1):
+            d[i, -1] = maxdist
+            d[i, 0] = i
+        for j in range(len(s2) + 1):
+            d[-1, j] = maxdist
+            d[0, j] = j
+
+        for i, cs1 in enumerate(s1, start=1):
+            db = 0
+            for j, cs2 in enumerate(s2, start=1):
+                i1 = da.get(cs2, 0)
+                j1 = db
+                if self.test_func(cs1, cs2):
+                    cost = 0
+                    db = j
+                else:
+                    cost = 1
+
+                d[i, j] = min(
+                    d[i - 1, j - 1] + cost,     # substitution
+                    d[i, j - 1] + 1,            # insertion
+                    d[i - 1, j] + 1,            # deletion
+                    d[i1 - 1, j1 - 1] + (i - i1) - 1 + (j - j1),  # transposition
+                )
+            da[cs1] = i
+
+        return d[len1, len2]
+
+    def _pure_python_restricted(self, s1: Sequence[T], s2: Sequence[T]) -> int:
         """
         https://www.guyrutenberg.com/2008/12/15/damerau-levenshtein-distance-in-python/
         """
-        d = {}
+        d: dict[tuple[int, int], int] = {}
 
         # matrix
         for i in range(-1, len(s1) + 1):
@@ -205,17 +278,19 @@ class DamerauLevenshtein(_Base):
 
         return d[len(s1) - 1, len(s2) - 1]
 
-    def __call__(self, s1, s2):
+    def __call__(self, s1: Sequence[T], s2: Sequence[T]) -> int:
         s1, s2 = self._get_sequences(s1, s2)
 
         result = self.quick_answer(s1, s2)
         if result is not None:
-            return result
+            return result  # type: ignore[return-value]
 
         # if numpy:
         #     return self._numpy(s1, s2)
         # else:
-        return self._pure_python(s1, s2)
+        if self.restricted:
+            return self._pure_python_restricted(s1, s2)
+        return self._pure_python_unrestricted(s1, s2)
 
 
 class JaroWinkler(_BaseSimilarity):
@@ -229,16 +304,23 @@ class JaroWinkler(_BaseSimilarity):
     https://github.com/Yomguithereal/talisman/blob/master/src/metrics/jaro.js
     https://github.com/Yomguithereal/talisman/blob/master/src/metrics/jaro-winkler.js
     """
-    def __init__(self, long_tolerance=False, winklerize=True, qval=1, external=True):
+
+    def __init__(
+        self,
+        long_tolerance: bool = False,
+        winklerize: bool = True,
+        qval: int = 1,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.long_tolerance = long_tolerance
         self.winklerize = winklerize
         self.external = external
 
-    def maximum(self, *sequences):
+    def maximum(self, *sequences: Sequence[object]) -> int:
         return 1
 
-    def __call__(self, s1, s2, prefix_weight=0.1):
+    def __call__(self, s1: Sequence[T], s2: Sequence[T], prefix_weight: float = 0.1) -> float:
         s1, s2 = self._get_sequences(s1, s2)
 
         result = self.quick_answer(s1, s2)
@@ -302,7 +384,7 @@ class JaroWinkler(_BaseSimilarity):
         # adjust for up to first 4 chars in common
         j = min(min_len, 4)
         i = 0
-        while i < j and s1[i] == s2[i] and s1[i]:
+        while i < j and s1[i] == s2[i]:
             i += 1
         if i:
             weight += i * prefix_weight * (1.0 - weight)
@@ -320,7 +402,12 @@ class JaroWinkler(_BaseSimilarity):
 
 
 class Jaro(JaroWinkler):
-    def __init__(self, long_tolerance=False, qval=1, external=True):
+    def __init__(
+        self,
+        long_tolerance: bool = False,
+        qval: int = 1,
+        external: bool = True,
+    ) -> None:
         super().__init__(
             long_tolerance=long_tolerance,
             winklerize=False,
@@ -341,9 +428,14 @@ class NeedlemanWunsch(_BaseSimilarity):
 
     https://en.wikipedia.org/wiki/Needleman%E2%80%93Wunsch_algorithm
     """
-    positive = False
 
-    def __init__(self, gap_cost=1.0, sim_func=None, qval=1, external=True):
+    def __init__(
+        self,
+        gap_cost: float = 1.0,
+        sim_func: SimFunc = None,
+        qval: int = 1,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.gap_cost = gap_cost
         if sim_func:
@@ -352,18 +444,18 @@ class NeedlemanWunsch(_BaseSimilarity):
             self.sim_func = self._ident
         self.external = external
 
-    def minimum(self, *sequences):
+    def minimum(self, *sequences: Sequence[object]) -> float:
         return -max(map(len, sequences)) * self.gap_cost
 
-    def maximum(self, *sequences):
+    def maximum(self, *sequences: Sequence[object]) -> float:
         return max(map(len, sequences))
 
-    def distance(self, *sequences):
+    def distance(self, *sequences: Sequence[object]) -> float:
         """Get distance between sequences
         """
         return -1 * self.similarity(*sequences)
 
-    def normalized_distance(self, *sequences):
+    def normalized_distance(self, *sequences: Sequence[object]) -> float:
         """Get distance from 0 to 1
         """
         minimum = self.minimum(*sequences)
@@ -372,8 +464,8 @@ class NeedlemanWunsch(_BaseSimilarity):
             return 0
         return (self.distance(*sequences) - minimum) / (maximum - minimum)
 
-    def normalized_similarity(self, *sequences):
-        """Get distance from 0 to 1
+    def normalized_similarity(self, *sequences: Sequence[object]) -> float:
+        """Get similarity from 0 to 1
         """
         minimum = self.minimum(*sequences)
         maximum = self.maximum(*sequences)
@@ -381,7 +473,7 @@ class NeedlemanWunsch(_BaseSimilarity):
             return 1
         return (self.similarity(*sequences) - minimum) / (maximum * 2)
 
-    def __call__(self, s1, s2):
+    def __call__(self, s1: Sequence[T], s2: Sequence[T]) -> float:
         if not numpy:
             raise ImportError('Please, install numpy for Needleman-Wunsch measure')
 
@@ -422,16 +514,23 @@ class SmithWaterman(_BaseSimilarity):
     https://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
     https://github.com/Yomguithereal/talisman/blob/master/src/metrics/smith-waterman.js
     """
-    def __init__(self, gap_cost=1.0, sim_func=None, qval=1, external=True):
+
+    def __init__(
+        self,
+        gap_cost: float = 1.0,
+        sim_func: SimFunc = None,
+        qval: int = 1,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.gap_cost = gap_cost
         self.sim_func = sim_func or self._ident
         self.external = external
 
-    def maximum(self, *sequences):
+    def maximum(self, *sequences: Sequence[object]) -> int:
         return min(map(len, sequences))
 
-    def __call__(self, s1, s2):
+    def __call__(self, s1: Sequence[T], s2: Sequence[T]) -> float:
         if not numpy:
             raise ImportError('Please, install numpy for Smith-Waterman measure')
 
@@ -464,7 +563,15 @@ class Gotoh(NeedlemanWunsch):
     penalties:
     https://www.cs.umd.edu/class/spring2003/cmsc838t/papers/gotoh1982.pdf
     """
-    def __init__(self, gap_open=1, gap_ext=0.4, sim_func=None, qval=1, external=True):
+
+    def __init__(
+        self,
+        gap_open: int = 1,
+        gap_ext: float = 0.4,
+        sim_func: SimFunc = None,
+        qval: int = 1,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.gap_open = gap_open
         self.gap_ext = gap_ext
@@ -474,13 +581,13 @@ class Gotoh(NeedlemanWunsch):
             self.sim_func = self._ident
         self.external = external
 
-    def minimum(self, *sequences):
+    def minimum(self, *sequences: Sequence[object]) -> int:
         return -min(map(len, sequences))
 
-    def maximum(self, *sequences):
+    def maximum(self, *sequences: Sequence[object]) -> int:
         return min(map(len, sequences))
 
-    def __call__(self, s1, s2):
+    def __call__(self, s1: Sequence[T], s2: Sequence[T]) -> float:
         if not numpy:
             raise ImportError('Please, install numpy for Gotoh measure')
 
@@ -536,7 +643,7 @@ class StrCmp95(_BaseSimilarity):
 
     http://cpansearch.perl.org/src/SCW/Text-JaroWinkler-0.1/strcmp95.c
     """
-    sp_mx = (
+    sp_mx: tuple[tuple[str, str], ...] = (
         ('A', 'E'), ('A', 'I'), ('A', 'O'), ('A', 'U'), ('B', 'V'), ('E', 'I'),
         ('E', 'O'), ('E', 'U'), ('I', 'O'), ('I', 'U'), ('O', 'U'), ('I', 'Y'),
         ('E', 'Y'), ('C', 'G'), ('E', 'F'), ('W', 'U'), ('W', 'V'), ('X', 'K'),
@@ -545,18 +652,18 @@ class StrCmp95(_BaseSimilarity):
         ('1', 'I'), ('1', 'L'), ('0', 'O'), ('0', 'Q'), ('C', 'K'), ('G', 'J'),
     )
 
-    def __init__(self, long_strings=False, external=True):
+    def __init__(self, long_strings: bool = False, external: bool = True) -> None:
         self.long_strings = long_strings
         self.external = external
 
-    def maximum(self, *sequences):
+    def maximum(self, *sequences: Sequence[object]) -> int:
         return 1
 
     @staticmethod
-    def _in_range(char):
+    def _in_range(char) -> bool:
         return 0 < ord(char) < 91
 
-    def __call__(self, s1, s2):
+    def __call__(self, s1: str, s2: str) -> float:
         s1 = s1.strip().upper()
         s2 = s2.strip().upper()
 
@@ -687,16 +794,22 @@ class MLIPNS(_BaseSimilarity):
     http://www.sial.iias.spb.su/files/386-386-1-PB.pdf
     https://github.com/Yomguithereal/talisman/blob/master/src/metrics/mlipns.js
     """
-    def __init__(self, threshold=0.25, maxmismatches=2, qval=1, external=True):
+
+    def __init__(
+        self, threshold: float = 0.25,
+        maxmismatches: int = 2,
+        qval: int = 1,
+        external: bool = True,
+    ) -> None:
         self.qval = qval
         self.threshold = threshold
         self.maxmismatches = maxmismatches
         self.external = external
 
-    def maximum(self, *sequences):
+    def maximum(self, *sequences: Sequence[object]) -> int:
         return 1
 
-    def __call__(self, *sequences):
+    def __call__(self, *sequences: Sequence[object]) -> float:
         sequences = self._get_sequences(*sequences)
 
         result = self.quick_answer(*sequences)
